/***********************************************************************
 ws-util.cpp - Some basic Winsock utility functions.

 This program is hereby released into the public domain.  There is
 ABSOLUTELY NO WARRANTY WHATSOEVER for this product.  Caveat hacker.
***********************************************************************/

#include <vcl.h>
#include "ws-util.h"

#include <iostream>
#include <algorithm>
#include <strstream>

//using namespace std;

#if !defined(_WINSOCK2API_)
// Winsock 2 header defines this, but Winsock 1.1 header doesn't.  In
// the interest of not requiring the Winsock 2 SDK which we don't really
// need, we'll just define this one constant ourselves.
#define SD_SEND 1
#endif

#define MAXHOSTNAME 254

//// Constants /////////////////////////////////////////////////////////

const int kBufferSize = 1024;


//// Statics ///////////////////////////////////////////////////////////

// List of Winsock error constants mapped to an interpretation string.
// Note that this list must remain sorted by the error constants'
// values, because we do a binary search on the list when looking up
// items.

static class ErrorEntry
{
public:
    int nID;
    const char* pcMessage;

    ErrorEntry(int id, const char* pc = 0) : nID(id), pcMessage(pc)
    {
    }

    bool operator<(const ErrorEntry& rhs)
    {
        return nID < rhs.nID;
    }

} gaErrorList[] =
  {
    ErrorEntry(0,                  "No error"),
    ErrorEntry(WSAEINTR,           "Interrupted system call"),
    ErrorEntry(WSAEBADF,           "Bad file number"),
    ErrorEntry(WSAEACCES,          "Permission denied"),
    ErrorEntry(WSAEFAULT,          "Bad address"),
    ErrorEntry(WSAEINVAL,          "Invalid argument"),
    ErrorEntry(WSAEMFILE,          "Too many open sockets"),
    ErrorEntry(WSAEWOULDBLOCK,     "Operation would block"),
    ErrorEntry(WSAEINPROGRESS,     "Operation now in progress"),
    ErrorEntry(WSAEALREADY,        "Operation already in progress"),
    ErrorEntry(WSAENOTSOCK,        "Socket operation on non-socket"),
    ErrorEntry(WSAEDESTADDRREQ,    "Destination address required"),
    ErrorEntry(WSAEMSGSIZE,        "Message too long"),
    ErrorEntry(WSAEPROTOTYPE,      "Protocol wrong type for socket"),
    ErrorEntry(WSAENOPROTOOPT,     "Bad protocol option"),
    ErrorEntry(WSAEPROTONOSUPPORT, "Protocol not supported"),
    ErrorEntry(WSAESOCKTNOSUPPORT, "Socket type not supported"),
    ErrorEntry(WSAEOPNOTSUPP,      "Operation not supported on socket"),
    ErrorEntry(WSAEPFNOSUPPORT,    "Protocol family not supported"),
    ErrorEntry(WSAEAFNOSUPPORT,    "Address family not supported"),
    ErrorEntry(WSAEADDRINUSE,      "Address already in use"),
    ErrorEntry(WSAEADDRNOTAVAIL,   "Can't assign requested address"),
    ErrorEntry(WSAENETDOWN,        "Network is down"),
    ErrorEntry(WSAENETUNREACH,     "Network is unreachable"),
    ErrorEntry(WSAENETRESET,       "Net connection reset"),
    ErrorEntry(WSAECONNABORTED,    "Software caused connection abort"),
    ErrorEntry(WSAECONNRESET,      "Connection reset by peer"),
    ErrorEntry(WSAENOBUFS,         "No buffer space available"),
    ErrorEntry(WSAEISCONN,         "Socket is already connected"),
    ErrorEntry(WSAENOTCONN,        "Socket is not connected"),
    ErrorEntry(WSAESHUTDOWN,       "Can't send after socket shutdown"),
    ErrorEntry(WSAETOOMANYREFS,    "Too many references, can't splice"),
    ErrorEntry(WSAETIMEDOUT,       "Connection timed out"),
    ErrorEntry(WSAECONNREFUSED,    "Connection refused"),
    ErrorEntry(WSAELOOP,           "Too many levels of symbolic links"),
    ErrorEntry(WSAENAMETOOLONG,    "File name too long"),
    ErrorEntry(WSAEHOSTDOWN,       "Host is down"),
    ErrorEntry(WSAEHOSTUNREACH,    "No route to host"),
    ErrorEntry(WSAENOTEMPTY,       "Directory not empty"),
    ErrorEntry(WSAEPROCLIM,        "Too many processes"),
    ErrorEntry(WSAEUSERS,          "Too many users"),
    ErrorEntry(WSAEDQUOT,          "Disc quota exceeded"),
    ErrorEntry(WSAESTALE,          "Stale NFS file handle"),
    ErrorEntry(WSAEREMOTE,         "Too many levels of remote in path"),
    ErrorEntry(WSASYSNOTREADY,     "Network system is unavailable"),
    ErrorEntry(WSAVERNOTSUPPORTED, "Winsock version out of range"),
    ErrorEntry(WSANOTINITIALISED,  "WSAStartup not yet called"),
    ErrorEntry(WSAEDISCON,         "Graceful shutdown in progress"),
    ErrorEntry(WSAHOST_NOT_FOUND,  "Host not found"),
    ErrorEntry(WSANO_DATA,         "No host data of that type was found")
};

bool operator<(const ErrorEntry& rhs1,const ErrorEntry& rhs2)
{
    return rhs1.nID < rhs2.nID;
}


const int kNumMessages = sizeof(gaErrorList) / sizeof(ErrorEntry);


//// WSAGetLastErrorMessage ////////////////////////////////////////////
// A function similar in spirit to Unix's perror() that tacks a canned
// interpretation of the value of WSAGetLastError() onto the end of a
// passed string, separated by a ": ".  Generally, you should implement
// smarter error handling than this, but for default cases and simple
// programs, this function is sufficient.
//
// This function returns a pointer to an internal static buffer, so you
// must copy the data from this function before you call it again.  It
// follows that this function is also not thread-safe.

const char* WSAGetLastErrorMessage(const char* pcMessagePrefix,
                                   int nErrorID /* = 0 */)
{
    // Build basic error string
    static char acErrorBuffer[256];
    std::ostrstream outs(acErrorBuffer, sizeof(acErrorBuffer));
    outs << pcMessagePrefix << ": ";

    // Tack appropriate canned message onto end of supplied message
    // prefix. Note that we do a binary search here: gaErrorList must be
	  // sorted by the error constant's value.
  	ErrorEntry* pEnd = gaErrorList + kNumMessages;
    ErrorEntry Target(nErrorID ? nErrorID : WSAGetLastError());
    ErrorEntry* it = std::lower_bound(gaErrorList, pEnd, Target);
    if ((it != pEnd) && (it->nID == Target.nID))
    {
        outs << it->pcMessage;
    }
    else
    {
        // Didn't find error in list, so make up a generic one
        outs << "unknown error";
    }
    outs << " (" << Target.nID << ")";

    // Finish error message off and return it.
    outs << std::ends;
    acErrorBuffer[sizeof(acErrorBuffer) - 1] = '\0';
    return acErrorBuffer;
}

// ---------------------------------------------------------------------------
int ConnectToHost(char *pHost, int iPort, bool bReturnErrorCode)
{
  int iSocket = INVALID_SOCKET;

  try
  {
    struct sockaddr_in sa;
    memset(&sa, 0, sizeof(struct sockaddr_in));

    struct hostent *hp = gethostbyname(pHost);
    if (hp == NULL)
    {
      AnsiString asSocketError = WSAGetLastErrorMessage("Calling gethostbyname()");
      throw EAbort(asSocketError);
    }

    sa.sin_family = AF_INET;                        // this is our host address
    sa.sin_addr   = *((LPIN_ADDR)*hp->h_addr_list); // this is our host address
    sa.sin_port   = htons(iPort);

    if ((iSocket = socket(PF_INET, SOCK_STREAM, 0)) < 0) /* create socket */
    {
      AnsiString asSocketError = WSAGetLastErrorMessage("Calling socket()");
      throw EAbort(asSocketError);
    }

    if(connect(iSocket,(struct sockaddr FAR*)&sa,sizeof(sa)) < 0)
    {
      AnsiString asSocketError = WSAGetLastErrorMessage("Calling connect()");
      throw EAbort(asSocketError);
    }
  }
  catch(EAbort &ex)
  {
    if(iSocket != INVALID_SOCKET)
    {
      closesocket(iSocket);
    }

    iSocket = INVALID_SOCKET;

    if(bReturnErrorCode == false)
    {
      throw EAbort(ex.Message);
    }
  }

  return(iSocket);
}

//// ShutdownConnection ////////////////////////////////////////////////
// Gracefully shuts the connection sd down.  Returns true if we're
// successful, false otherwise.
// ---------------------------------------------------------------------------
bool ShutdownConnection(SOCKET sd, bool bReceiveLastPacket)
{
  bool bRet = true;

  if(sd == INVALID_SOCKET)
  {
    bRet = false;
  }
  else
  {
    // Disallow any further data sends.  This will tell the other side
    // that we want to go away now.  If we skip this step, we don't
    // shut the connection down nicely.
    if (shutdown(sd, SD_SEND) == SOCKET_ERROR)
    {
      bRet = false;
    }

    if(bReceiveLastPacket == true)
    {
      // Receive any extra data still sitting on the socket.  After all
      // data is received, this call will block until the remote host
      // acknowledges the TCP control packet sent by the shutdown above.
      // Then we'll get a 0 back from recv, signalling that the remote
      // host has closed its side of the connection.
      char acReadBuffer[kBufferSize] = "";
      while (1)
      {
          int nNewBytes = recv(sd, acReadBuffer, kBufferSize, 0);
          if (nNewBytes == SOCKET_ERROR)
          {
              bRet = false;
              break;
          }
          else if (nNewBytes != 0)
          {
              std::cerr << std::endl << "FYI, received " << nNewBytes <<
                      " unexpected bytes during shutdown." << std::endl;
          }
          else
          {
              // Okay, we're done!
              break;
          }
      }
    }

    // Close the socket.
    if (closesocket(sd) == SOCKET_ERROR)
    {
        bRet = false;
    }
  }

  return bRet;
}

// ---------------------------------------------------------------------------
void SetupSocketMonitoring(int &iSocket)
{
  struct timeval tv;
  tv.tv_sec   = 0;
  tv.tv_usec  = 0;

  fd_set rfds;
  FD_ZERO(&rfds);
  FD_SET(iSocket, &rfds);

  int iFound = select(0, &rfds, NULL, NULL, &tv);

  // A negative value indicates some type of error
  if(iFound == SOCKET_ERROR)
  {
    //int iSocketErr = WSAGetLastError();
    //const char *pError = WSAGetLastErrorMessage("Calling select: ");
    AnsiString asError = WSAGetLastErrorMessage("Calling SetupSocketMonitoring: ");
    throw Exception(asError);
  }
  else if(iFound > 0)
  {
    if(FD_ISSET(iSocket, &rfds) != 0)
    {

    }
  }
}

bool CheckForSocketData(int &iSocket,fd_set *fdperm_rfds, struct timeval *pTimeValue)
{
  bool bData = false;

  if(iSocket > 0)
  {
    fd_set rfds;
    struct timeval tv;

    if(fdperm_rfds == NULL)
    {
      FD_ZERO(&rfds);
      FD_SET(iSocket, &rfds);
      //pFD = &rfds;
    }
    else
    {
      //memcpy(&rfds,fdperm_rfds,sizeof(fdperm_rfds));
      rfds = *fdperm_rfds;
    }

    if(pTimeValue == NULL)
    {
      tv.tv_sec   = 1;
      tv.tv_usec  = 0;

      pTimeValue = &tv;
    }

    int iFound = select(0, &rfds, NULL, NULL, pTimeValue);

    // A negative value indicates some type of error
    if(iFound == SOCKET_ERROR)
    {
      //int iSocketErr = WSAGetLastError();
      //const char *pError = WSAGetLastErrorMessage("Calling select: ");
      AnsiString asError = WSAGetLastErrorMessage("Calling CheckForSocketData (select): ");
      throw Exception(asError);
    }
    else if(iFound > 0)
    {
      if(FD_ISSET(iSocket, &rfds) > 0)
      {
          bData = true;
      }
    }

  //  if(fdperm_rfds != NULL)
  //  {
  //    FD_ZERO(fdperm_rfds);
  //    FD_SET(iSocket, fdperm_rfds);
  //  }
  }

  return(bData);
}

// ---------------------------------------------------------------------------
int ListenForSocketConnection(int iPort)
{
  struct sockaddr_in sa;
  memset(&sa, 0, sizeof(struct sockaddr_in)); /* clear our address */

  char   szMyHostName[MAXHOSTNAME + 1]  = "";
  int iRetval = gethostname(szMyHostName, MAXHOSTNAME);           /* who are we? */

  if(iRetval == SOCKET_ERROR)
  {
    AnsiString asSocketError = WSAGetLastErrorMessage("Calling gethostname()");
    throw EAbort(asSocketError);
  }

  //printf("Running on hostname: [%s]\n",myname);

  struct hostent *hp = gethostbyname(szMyHostName); // get our address info
  if (hp == NULL)                                   // we don't exist !?
  {
    AnsiString asSocketError = WSAGetLastErrorMessage("Calling gethostbyname()");
    throw EAbort(asSocketError);
  }

  //printf("IP Addresses used on this PC\r\n");
  /*
  for(int i = 0; hp->h_addr_list[i] != NULL; i++)
  {
    // inet_ntoa converts the ints to a string also.
    printf("IP address: [%d\.%d\.%d\.%d]\n",(int)(unsigned char)hp->h_addr_list[i][0],
                                            (int)(unsigned char)hp->h_addr_list[i][1],
                                            (int)(unsigned char)hp->h_addr_list[i][2],
                                            (int)(unsigned char)hp->h_addr_list[i][3]);
  }
  */

  sa.sin_family       = hp->h_addrtype;              // this is our host address
  sa.sin_port         = htons(iPort);                // this is our port number
  sa.sin_addr.s_addr  = htonl(INADDR_ANY);           // Bind to all IP Addresses

  int iSocket = INVALID_SOCKET;
  if ((iSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0) /* create socket */
  {
    AnsiString asSocketError = WSAGetLastErrorMessage("Calling socket()");
    throw EAbort(asSocketError);
  }

  char opttrue = 1;
  //char optfalse = 0;
  setsockopt(iSocket,SOL_SOCKET,SO_REUSEADDR,&opttrue,sizeof(opttrue));

  if (bind(iSocket,(struct sockaddr *)&sa,sizeof(struct sockaddr_in)) < 0)
  {
    closesocket(iSocket);

    AnsiString asSocketError = WSAGetLastErrorMessage("Calling bind()");
    throw EAbort(asSocketError);
  }

  if (listen(iSocket, 100) < 0) // max # of queued connects
  {
    closesocket(iSocket);

    AnsiString asSocketError = WSAGetLastErrorMessage("Calling listen()");
    throw EAbort(asSocketError);
  }

  return(iSocket);
}

// ---------------------------------------------------------------------------
int AcceptClientConnection(int iSocket,sockaddr_in *pSocketAddress)
{
  struct sockaddr_in tSocketAddress;
  if(pSocketAddress == NULL)
  {
    pSocketAddress = &tSocketAddress;
  }

  memset(pSocketAddress, 0, sizeof(struct sockaddr_in));

  int iSize = sizeof(*pSocketAddress);

  int iNewSocket = -1;
  if ((iNewSocket = accept(iSocket,(struct sockaddr *)pSocketAddress,&iSize)) < 0)   // accept connection if there is one
  {
    AnsiString asSocketError = WSAGetLastErrorMessage("Calling accept()");
    throw EAbort(asSocketError);
  }

  /*
  printf("Client IP Address [%d.%d.%d.%d] for SD: [%d]\r\n",
          (int)sa.sin_addr.S_un.S_un_b.s_b1, (int)sa.sin_addr.S_un.S_un_b.s_b2,
          (int)sa.sin_addr.S_un.S_un_b.s_b3, (int)sa.sin_addr.S_un.S_un_b.s_b4,t);
  */

  return(iNewSocket);
}

// ---------------------------------------------------------------------------
